/**
 * Inject functions
 * @function inject
 * @param {function} mapper - Props mapper
 * @returns {function} Wrap function
 */
'use strict'

const React = require('react')
const { unlessProduction } = require('the-check')
const contexts = require('./contexts')
const { createElement: c } = React
const { HandleContext } = contexts

/** @lends inject */
function inject (mapper) {
  return function wrap (Component, options = {}) {
    const {
      catcher = function invokeCatcher (e) {
        if (e.resolved) {
        } else {
          throw e
        }
      },
    } = options

    class Injected extends React.Component {
      constructor (props) {
        super(props)
        this.state = { mapped: {} }
      }

      componentWillMount () {
        const mapped = this.mapMethods()
        this.setState({ mapped })
      }

      mapMethods () {
        const { handle = this.props.handle } = this.context
        const values = { ...this.props, ...this.context }
        const trapped = this
        const propsProxy = new Proxy(
          values,
          {
            get: (target, name) => {
              const fromProp = trapped.props[name]
              if (typeof fromProp !== 'undefined') {
                return fromProp
              }
              const fromContext = trapped.context[name]
              if (typeof fromContext !== 'undefined') {
                return fromContext
              }
              unlessProduction(() => {
                const ComponentName = (Component.wrapStack || []).reverse()[0] || Component.displayName || Component.name
                console.warn(`[${ComponentName}] "${name}" is not found in propsProxy`)
              })
              return void(0)
            },
            set: (target, name, value) => {
              throw new Error('[TheHandle] Can not set value')
            },
          })
        if (!handle) {
          throw new Error(`[TheHandle] Handle not found`)
        }
        const mapped = Object.assign({}, mapper(handle, propsProxy, this.props, this.context))
        for (const [name, value] of Object.entries(mapped)) {
          if (typeof value === 'function') {
            mapped[name] = function injectedFunctionWrap (...args) {
              try {
                const result = value.call(mapped, ...args)
                const isPromise = !!(result && result.then && result.catch)
                if (isPromise) {
                  return result.catch(catcher)
                }
                return result
              } catch (e) {
                catcher(e)
              }
            }
          }
        }
        return mapped
      }

      render () {
        const { props, state } = this
        return c(
          Component,
          { ...props, ...state.mapped },
        )
      }
    }

    Injected.contextType = HandleContext

    return Injected
  }
}

module.exports = inject
